from ninja import Schema
from ninja.router import Router

from ninja_jwt.schema_control import SchemaControl
from ninja_jwt.settings import api_settings

blacklist_router = Router()

schema = SchemaControl(api_settings)


@blacklist_router.post(
    "/blacklist",
    response={200: Schema},
    url_name="token_blacklist",
    auth=None,
)
def blacklist_token(request, refresh: schema.blacklist_schema):
    return refresh.to_response_schema()



import typing
import warnings
from typing import Any, Dict, Optional, Type, cast

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import AbstractUser, update_last_login
from django.utils.translation import gettext_lazy as _
from ninja import ModelSchema, Schema
from ninja.schema import DjangoGetter
from pydantic import model_validator

import ninja_jwt.exceptions as exceptions
from ninja_jwt.utils import token_error

from .settings import api_settings
from .tokens import RefreshToken, SlidingToken, UntypedToken

if api_settings.BLACKLIST_AFTER_ROTATION:
    from .token_blacklist.models import BlacklistedToken

user_name_field = get_user_model().USERNAME_FIELD  # type: ignore


class AuthUserSchema(ModelSchema):
    class Config:
        model = get_user_model()
        model_fields = [user_name_field]


class InputSchemaMixin:
    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        raise NotImplementedError("Must implement `get_response_schema`")

    def to_response_schema(self):
        _schema_type = self.get_response_schema()
        return _schema_type(**self.model_dump())


class TokenInputSchemaMixin(InputSchemaMixin):
    _user: Optional[AbstractUser] = None

    _default_error_messages = {
        "no_active_account": _("No active account found with the given credentials")
    }

    def check_user_authentication_rule(self) -> None:
        if not api_settings.USER_AUTHENTICATION_RULE(self._user):
            raise exceptions.AuthenticationFailed(
                self._default_error_messages["no_active_account"]
            )

    @classmethod
    def validate_values(cls, values: Dict) -> Dict:
        if user_name_field not in values and "password" not in values:
            raise exceptions.ValidationError(
                {
                    user_name_field: f"{user_name_field} is required",
                    "password": "password is required",
                }
            )

        if not values.get(user_name_field):
            raise exceptions.ValidationError(
                {user_name_field: f"{user_name_field} is required"}
            )

        if not values.get("password"):
            raise exceptions.ValidationError({"password": "password is required"})

        _user = authenticate(**values)
        cls._user = _user

        if not (_user is not None and _user.is_active):
            raise exceptions.AuthenticationFailed(
                cls._default_error_messages["no_active_account"]
            )

        return values

    def output_schema(self) -> Schema:
        warnings.warn(
            "output_schema() is deprecated in favor of " "to_response_schema()",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.to_response_schema()

    @classmethod
    def get_token(cls, user: AbstractUser) -> Dict:
        raise NotImplementedError(
            "Must implement `get_token` method for `TokenObtainSerializer` subclasses"
        )


class TokenObtainInputSchemaBase(ModelSchema, TokenInputSchemaMixin):
    class Config:
        # extra = "allow"
        model = get_user_model()
        model_fields = ["password", user_name_field]

    @model_validator(mode="before")
    def validate_inputs(cls, values: DjangoGetter) -> DjangoGetter:
        input_values = values._obj
        if isinstance(input_values, dict):
            values._obj.update(cls.validate_values(input_values))
            return values
        return values

    @model_validator(mode="after")
    def post_validate(cls, values: Dict) -> dict:
        return cls.post_validate_schema(values)

    @classmethod
    def post_validate_schema(cls, values: Dict) -> dict:
        """
        This is a post validate process which is common for any token generating schema.
        :param values:
        :return:
        """
        # get_token can return values that wants to apply to `OutputSchema`

        data = cls.get_token(cls._user)

        if not isinstance(data, dict):
            raise Exception("`get_token` must return a `typing.Dict` type.")

        # a workaround for extra attributes since adding extra=allow in modelconfig adds `addition_props`
        # field to the schema
        values.__dict__.update(token_data=data)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, cls._user)

        return values

    def get_response_schema_init_kwargs(self) -> dict:
        return dict(
            self.dict(exclude={"password"}), **self.__dict__.get("token_data", {})
        )

    def to_response_schema(self):
        _schema_type = self.get_response_schema()
        return _schema_type(**self.get_response_schema_init_kwargs())


class TokenObtainPairOutputSchema(AuthUserSchema):
    refresh: str
    access: str


class TokenObtainPairInputSchema(TokenObtainInputSchemaBase):
    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        return TokenObtainPairOutputSchema

    @classmethod
    def get_token(cls, user: AbstractUser) -> Dict:
        values = {}
        refresh = RefreshToken.for_user(user)
        refresh = cast(RefreshToken, refresh)
        values["refresh"] = str(refresh)
        values["access"] = str(refresh.access_token)
        return values


class TokenObtainSlidingOutputSchema(AuthUserSchema):
    token: str


class TokenObtainSlidingInputSchema(TokenObtainInputSchemaBase):
    @classmethod
    def get_response_schema(cls) -> Type:
        return TokenObtainSlidingOutputSchema

    @classmethod
    def get_token(cls, user: AbstractUser) -> Dict:
        values = {}
        slide_token = SlidingToken.for_user(user)
        values["token"] = str(slide_token)
        return values


class TokenRefreshInputSchema(Schema, InputSchemaMixin):
    refresh: str

    @model_validator(mode="before")
    def validate_schema(cls, values: DjangoGetter) -> dict:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("refresh"):
                raise exceptions.ValidationError({"refresh": "token is required"})
        return values

    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        return TokenRefreshOutputSchema


class TokenRefreshOutputSchema(Schema):
    refresh: str
    access: Optional[str]

    @model_validator(mode="before")
    @token_error
    def validate_schema(cls, values: DjangoGetter) -> typing.Any:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("refresh"):
                raise exceptions.ValidationError(
                    {"refresh": "refresh token is required"}
                )

            refresh = RefreshToken(values["refresh"])

            data = {"access": str(refresh.access_token)}

            if api_settings.ROTATE_REFRESH_TOKENS:
                if api_settings.BLACKLIST_AFTER_ROTATION:
                    try:
                        # Attempt to blacklist the given refresh token
                        refresh.blacklist()
                    except AttributeError:
                        # If blacklist app not installed, `blacklist` method will
                        # not be present
                        pass

                refresh.set_jti()
                refresh.set_exp()
                refresh.set_iat()

                data["refresh"] = str(refresh)
            values.update(data)
        return values


class TokenRefreshSlidingInputSchema(Schema, InputSchemaMixin):
    token: str

    @model_validator(mode="before")
    def validate_schema(cls, values: DjangoGetter) -> dict:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("token"):
                raise exceptions.ValidationError({"token": "token is required"})
        return values

    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        return TokenRefreshSlidingOutputSchema


class TokenRefreshSlidingOutputSchema(Schema):
    token: str

    @model_validator(mode="before")
    @token_error
    def validate_schema(cls, values: DjangoGetter) -> dict:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("token"):
                raise exceptions.ValidationError({"token": "token is required"})

            token = SlidingToken(values["token"])

            # Check that the timestamp in the "refresh_exp" claim has not
            # passed
            token.check_exp(api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM)

            # Update the "exp" and "iat" claims
            token.set_exp()
            token.set_iat()
            values.update({"token": str(token)})
        return values


class TokenVerifyInputSchema(Schema, InputSchemaMixin):
    token: str

    @model_validator(mode="before")
    @token_error
    def validate_schema(cls, values: DjangoGetter) -> Dict:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("token"):
                raise exceptions.ValidationError({"token": "token is required"})
            token = UntypedToken(values["token"])

            if (
                api_settings.BLACKLIST_AFTER_ROTATION
                and "ninja_jwt.token_blacklist" in settings.INSTALLED_APPS
            ):
                jti = token.get(api_settings.JTI_CLAIM)
                if BlacklistedToken.objects.filter(token__jti=jti).exists():
                    raise exceptions.ValidationError("Token is blacklisted")

        return values

    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        return Schema

    def to_response_schema(self):
        return {}


class TokenBlacklistInputSchema(Schema, InputSchemaMixin):
    refresh: str

    @model_validator(mode="before")
    @token_error
    def validate_schema(cls, values: DjangoGetter) -> dict:
        values = values._obj

        if isinstance(values, dict):
            if not values.get("refresh"):
                raise exceptions.ValidationError(
                    {"refresh": "refresh token is required"}
                )
            refresh = RefreshToken(values["refresh"])
            try:
                refresh.blacklist()
            except AttributeError:
                pass
        return values

    @classmethod
    def get_response_schema(cls) -> Type[Schema]:
        return Schema

    def to_response_schema(self):
        return {}


__deprecated__ = {
    "TokenBlacklistSerializer": TokenBlacklistInputSchema,
    "TokenVerifySerializer": TokenVerifyInputSchema,
    "TokenRefreshSlidingSerializer": TokenRefreshSlidingOutputSchema,
    "TokenRefreshSlidingSchema": TokenRefreshSlidingInputSchema,
    "TokenRefreshSerializer": TokenRefreshOutputSchema,
    "TokenRefreshSchema": TokenRefreshInputSchema,
    "TokenObtainSlidingOutput": TokenObtainSlidingOutputSchema,
    "TokenObtainSerializer": TokenObtainInputSchemaBase,
    "TokenObtainPairOutput": TokenObtainPairOutputSchema,
    "TokenObtainPairSerializer": TokenObtainPairInputSchema,
    "TokenObtainSlidingSerializer": TokenObtainSlidingInputSchema,
}


def __getattr__(name: str) -> Any:  # pragma: no cover
    if name in __deprecated__:
        value = __deprecated__[name]
        warnings.warn(
            f"'{name}' is deprecated. Use '{value.__name__}' instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return value
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from ninja import Schema
from ninja.router import Router

from ninja_jwt.schema_control import SchemaControl
from ninja_jwt.settings import api_settings

schema = SchemaControl(api_settings)

verify_router = Router()


@verify_router.post(
    "/verify",
    response={200: Schema},
    url_name="token_verify",
    auth=None,
)
def verify_token(request, token: schema.verify_schema):
    return token.to_response_schema()



from ninja.router import Router

from ninja_jwt.schema_control import SchemaControl
from ninja_jwt.settings import api_settings

schema = SchemaControl(api_settings)

obtain_pair_router = Router()
sliding_router = Router()


@obtain_pair_router.post(
    "/pair",
    response=schema.obtain_pair_schema.get_response_schema(),
    url_name="token_obtain_pair",
    auth=None,
)
def obtain_token(request, user_token: schema.obtain_pair_schema):
    user_token.check_user_authentication_rule()
    return user_token.output_schema()


@obtain_pair_router.post(
    "/refresh",
    response=schema.obtain_pair_refresh_schema.get_response_schema(),
    url_name="token_refresh",
    auth=None,
)
def refresh_token(request, refresh_token: schema.obtain_pair_refresh_schema):
    return refresh_token.to_response_schema()


@sliding_router.post(
    "/sliding",
    response=schema.obtain_sliding_schema.get_response_schema(),
    url_name="token_obtain_sliding",
    auth=None,
)
def obtain_token_sliding_token(request, user_token: schema.obtain_sliding_schema):
    user_token.check_user_authentication_rule()
    return user_token.to_response_schema()


@sliding_router.post(
    "/sliding/refresh",
    response=schema.obtain_sliding_refresh_schema.get_response_schema(),
    url_name="token_refresh_sliding",
    auth=None,
)
def refresh_token_sliding(request, refresh_token: schema.obtain_sliding_refresh_schema):
    return refresh_token.to_response_schema()



import logging
from calendar import timegm
from datetime import datetime
from functools import wraps
from importlib import import_module

from django.conf import settings
from django.utils.functional import lazy
from django.utils.timezone import is_naive, make_aware

from ninja_jwt import exceptions

try:
    from datetime import timezone
except ImportError:
    from django.utils import timezone


logger = logging.getLogger("django")


def token_error(func):
    @wraps(func)
    def _wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exceptions.TokenError as tex:
            raise exceptions.InvalidToken(str(tex)) from tex
        except Exception as ex:
            logger.error(f"{func} raised exception: {str(ex)}")
            raise ex

    return _wrap


def import_callable(path_or_callable):
    if callable(path_or_callable):
        return path_or_callable
    else:
        assert isinstance(path_or_callable, str)
        package, attr = path_or_callable.rsplit(".", 1)
        packages = import_module(package)
        return getattr(packages, attr)


def make_utc(dt):
    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt, timezone=timezone.utc)

    return dt


def aware_utcnow():
    dt = datetime.now(tz=timezone.utc)
    if not settings.USE_TZ:
        dt = dt.replace(tzinfo=None)

    return dt


def datetime_to_epoch(dt):
    return timegm(dt.utctimetuple())


def datetime_from_epoch(ts):
    dt = datetime.fromtimestamp(ts, tz=timezone.utc)
    if not settings.USE_TZ:
        dt = dt.replace(tzinfo=None)

    return dt


def format_lazy(s, *args, **kwargs):
    return s.format(*args, **kwargs)


format_lazy = lazy(format_lazy, str)



from asgiref.sync import sync_to_async
from ninja import Schema
from ninja_extra import ControllerBase, api_controller, http_post
from ninja_extra.permissions import AllowAny

from ninja_jwt.schema_control import SchemaControl
from ninja_jwt.settings import api_settings

__all__ = [
    "TokenVerificationController",
    "TokenBlackListController",
    "TokenObtainPairController",
    "TokenObtainSlidingController",
    "TokenObtainSlidingController",
    "NinjaJWTDefaultController",
    "NinjaJWTSlidingController",
    "AsyncTokenVerificationController",
    "AsyncTokenBlackListController",
    "AsyncTokenObtainPairController",
    "AsyncTokenObtainSlidingController",
    "AsyncTokenObtainSlidingController",
    "AsyncNinjaJWTDefaultController",
    "AsyncNinjaJWTSlidingController",
]

schema = SchemaControl(api_settings)


class TokenVerificationController(ControllerBase):
    auto_import = False

    @http_post(
        "/verify",
        response={200: Schema},
        url_name="token_verify",
    )
    def verify_token(self, token: schema.verify_schema):
        return token.to_response_schema()


class TokenBlackListController(ControllerBase):
    auto_import = False

    @http_post(
        "/blacklist",
        response={200: Schema},
        url_name="token_blacklist",
    )
    def blacklist_token(self, refresh: schema.blacklist_schema):
        return refresh.to_response_schema()


class TokenObtainPairController(ControllerBase):
    auto_import = False

    @http_post(
        "/pair",
        response=schema.obtain_pair_schema.get_response_schema(),
        url_name="token_obtain_pair",
    )
    def obtain_token(self, user_token: schema.obtain_pair_schema):
        user_token.check_user_authentication_rule()
        return user_token.to_response_schema()

    @http_post(
        "/refresh",
        response=schema.obtain_pair_refresh_schema.get_response_schema(),
        url_name="token_refresh",
    )
    def refresh_token(self, refresh_token: schema.obtain_pair_refresh_schema):
        return refresh_token.to_response_schema()


class TokenObtainSlidingController(TokenObtainPairController):
    auto_import = False

    @http_post(
        "/sliding",
        response=schema.obtain_sliding_schema.get_response_schema(),
        url_name="token_obtain_sliding",
    )
    def obtain_token(self, user_token: schema.obtain_sliding_schema):
        user_token.check_user_authentication_rule()
        return user_token.to_response_schema()

    @http_post(
        "/sliding/refresh",
        response=schema.obtain_sliding_refresh_schema.get_response_schema(),
        url_name="token_refresh_sliding",
    )
    def refresh_token(self, refresh_token: schema.obtain_sliding_refresh_schema):
        return refresh_token.to_response_schema()


@api_controller("/token", permissions=[AllowAny], tags=["token"], auth=None)
class NinjaJWTDefaultController(TokenVerificationController, TokenObtainPairController):
    """NinjaJWT Default controller for obtaining and refreshing tokens"""

    auto_import = False


@api_controller("/token", permissions=[AllowAny], tags=["token"], auth=None)
class NinjaJWTSlidingController(
    TokenVerificationController, TokenObtainSlidingController
):
    """
    NinjaJWT Sliding controller for obtaining and refreshing tokens
    Add 'ninja_jwt.tokens.SlidingToken' in AUTH_TOKEN_CLASSES in Settings
    """

    auto_import = False


class AsyncTokenVerificationController(TokenVerificationController):
    @http_post(
        "/verify",
        response={200: Schema},
        url_name="token_verify",
    )
    async def verify_token(self, token: schema.verify_schema):
        return token.to_response_schema()


class AsyncTokenBlackListController(TokenBlackListController):
    auto_import = False

    @http_post(
        "/blacklist",
        response={200: Schema},
        url_name="token_blacklist",
    )
    async def blacklist_token(self, refresh: schema.blacklist_schema):
        return refresh.to_response_schema()


class AsyncTokenObtainPairController(TokenObtainPairController):
    @http_post(
        "/pair",
        response=schema.obtain_pair_schema.get_response_schema(),
        url_name="token_obtain_pair",
    )
    async def obtain_token(self, user_token: schema.obtain_pair_schema):
        await sync_to_async(user_token.check_user_authentication_rule)()
        return user_token.to_response_schema()

    @http_post(
        "/refresh",
        response=schema.obtain_pair_refresh_schema.get_response_schema(),
        url_name="token_refresh",
    )
    async def refresh_token(self, refresh_token: schema.obtain_pair_refresh_schema):
        refresh = await sync_to_async(refresh_token.to_response_schema)()
        return refresh


class AsyncTokenObtainSlidingController(TokenObtainSlidingController):
    @http_post(
        "/sliding",
        response=schema.obtain_sliding_schema.get_response_schema(),
        url_name="token_obtain_sliding",
    )
    async def obtain_token(self, user_token: schema.obtain_sliding_schema):
        await sync_to_async(user_token.check_user_authentication_rule)()
        return user_token.to_response_schema()

    @http_post(
        "/sliding/refresh",
        response=schema.obtain_sliding_refresh_schema.get_response_schema(),
        url_name="token_refresh_sliding",
    )
    async def refresh_token(self, refresh_token: schema.obtain_sliding_refresh_schema):
        refresh = await sync_to_async(refresh_token.to_response_schema)()
        return refresh


@api_controller("/token", permissions=[AllowAny], tags=["token"], auth=None)
class AsyncNinjaJWTDefaultController(
    AsyncTokenVerificationController, AsyncTokenObtainPairController
):
    """NinjaJWT Async Default controller for obtaining and refreshing tokens"""

    auto_import = False


@api_controller("/token", permissions=[AllowAny], tags=["token"], auth=None)
class AsyncNinjaJWTSlidingController(
    AsyncTokenVerificationController, AsyncTokenObtainSlidingController
):
    """
    NinjaJWT Async Sliding controller for obtaining and refreshing tokens
    Add 'ninja_jwt.tokens.SlidingToken' in AUTH_TOKEN_CLASSES in Settings
    """

    auto_import = False



from typing import TYPE_CHECKING, Type

from django.utils.module_loading import import_string

from ninja_jwt.schema import InputSchemaMixin, TokenInputSchemaMixin

if TYPE_CHECKING:  # pragma: no cover
    from ninja_jwt.settings import NinjaJWTSettings


class SchemaControl:
    """
    A Schema Helper Class that imports Schema from configurations
    """

    def __init__(self, api_settings: "NinjaJWTSettings") -> None:
        self._verify_schema = import_string(api_settings.TOKEN_VERIFY_INPUT_SCHEMA)
        self.validate_type(
            self._verify_schema, InputSchemaMixin, "TOKEN_VERIFY_INPUT_SCHEMA"
        )

        self._blacklist_schema = import_string(
            api_settings.TOKEN_BLACKLIST_INPUT_SCHEMA
        )
        self.validate_type(
            self._blacklist_schema, InputSchemaMixin, "TOKEN_BLACKLIST_INPUT_SCHEMA"
        )

        self._obtain_pair_schema = import_string(
            api_settings.TOKEN_OBTAIN_PAIR_INPUT_SCHEMA
        )
        self.validate_type(
            self._obtain_pair_schema,
            TokenInputSchemaMixin,
            "TOKEN_OBTAIN_PAIR_INPUT_SCHEMA",
        )

        self._obtain_pair_refresh_schema = import_string(
            api_settings.TOKEN_OBTAIN_PAIR_REFRESH_INPUT_SCHEMA
        )

        self.validate_type(
            self._obtain_pair_refresh_schema,
            InputSchemaMixin,
            "TOKEN_OBTAIN_PAIR_REFRESH_INPUT_SCHEMA",
        )

        self._obtain_sliding_schema = import_string(
            api_settings.TOKEN_OBTAIN_SLIDING_INPUT_SCHEMA
        )
        self.validate_type(
            self._obtain_sliding_schema,
            TokenInputSchemaMixin,
            "TOKEN_OBTAIN_SLIDING_INPUT_SCHEMA",
        )

        self._obtain_sliding_refresh_schema = import_string(
            api_settings.TOKEN_OBTAIN_SLIDING_REFRESH_INPUT_SCHEMA
        )
        self.validate_type(
            self._obtain_sliding_refresh_schema,
            InputSchemaMixin,
            "TOKEN_OBTAIN_SLIDING_REFRESH_INPUT_SCHEMA",
        )

    def validate_type(
        self, schema_type: Type, sub_class: Type, settings_key: str
    ) -> None:
        if not issubclass(schema_type, sub_class):
            raise Exception(f"{settings_key} type must inherit from `{sub_class}`")

    @property
    def verify_schema(self) -> "TokenInputSchemaMixin":
        return self._verify_schema

    @property
    def blacklist_schema(self) -> "TokenInputSchemaMixin":
        return self._blacklist_schema

    @property
    def obtain_pair_schema(self) -> "TokenInputSchemaMixin":
        return self._obtain_pair_schema

    @property
    def obtain_pair_refresh_schema(self) -> "TokenInputSchemaMixin":
        return self._obtain_pair_refresh_schema

    @property
    def obtain_sliding_schema(self) -> "TokenInputSchemaMixin":
        return self._obtain_sliding_schema

    @property
    def obtain_sliding_refresh_schema(self) -> "TokenInputSchemaMixin":
        return self._obtain_sliding_refresh_schema
